#!/usr/bin/python
# TS6 Pseudoserver plugin for PyPseudoserver
#todo:
#re-design server/user hashes
#export command prefix to config file

from pseudoprotocol import PseudoProtocol
import time, datetime, re, os
import networkx as nx
from __main__ import *

class TS6Protocol(PseudoProtocol):
	"""A PseudoServer based on the TS6 IRCD protocol
	
	Extensions to PseudoProtocol:
	[my_]user_t has the following extra fields:
	KEY = UID
	'ts' = timestamp of the user's signon
	'sid' = TS6 SID of the server the user is on
	'flags' = ENCAP AUTHFLAGS (authflags) [default=""]
	'su' = ENCAP SU (services user) [default=""]
	
	server_t has the following extra fields:
	KEY = SID
	'uplink_sid' = TS6 SID of the uplink of the server
	"""
	
	version = u"%s.TS6" % PseudoProtocol.version
	sid = config.get('server', 'sid')
	uplink_sid = ""
	burst_timer = 0
	uid = ""
	my_client_count = 0 # lifetime connections
	use_uid = True
	flood_noticed = False
	silence = False

	prefix = config.get('control', 'prefix')

	def __init__(self):
		"""Initialize various components"""
		PseudoProtocol.__init__(self)
		self.server_tree.add_node(self.sid)

		self.sjoinresub = re.compile(u"[@+~%&]").sub
		
### Begin helper methods	
	def base36encode(self, input):
		"""Takes an integer and returns it as base36 encoded
		"""
		
		CLIST="ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
		rv = ""
		while input != 0:
			rv = CLIST[input % 36] + rv
			input /= 36
		rv = "%6s" % rv
		return rv.replace(" ", "A")
	
	def get_user(self, ident):
		"""Designed to superceed common use of the old name_to_*id/*id_to_name
		functions.  This gets a user dict regardless of it being
		a uid or name supplied.

		Optimized for speed, as it is commonly used
		Assumes most calls are with UID
		"""
		
		try:
			return (ident, self.user_t[ident])
		except KeyError:
			lident = ident.lower()
			lower = unicode.lower
			for k,v in self.user_t.iteritems():
				if lower(v['nick']) == lident: return (k, v)
			return (None, None)

	def get_server(self, ident):
		"""Designed to superceed common use of the old name_to_*id/*id_to_name
		functions.  This gets a server dict regardless of it being
		a sid or name supplied.

		Optimized for speed
		Assumes most calls are with SID
		"""
		
		try:
			return (ident, self.server_t[ident])
		except KeyError:		
			lident = ident.lower()
			lower = unicode.lower
			for k,v in self.server_t.iteritems():
				if lower(v['host']) == lident: return (k, v)
			return (None, None)

		
	def get_channel(self, ident, skip_loop=False):
		"""Returns a pointer to a channel descriptor given its name
		
		Optimized for speed
		Assumes most calls are with proper channel case
		"""
		
		try:
			return (ident, self.channel_t[ident])
		except KeyError:
			# possibility its a case-mismatch scenario
			lident = ident.lower()
			lower = unicode.lower
			for k,v in self.channel_t.iteritems():
				if lower(k) == lident: return (k, v)
			return (None, None)
		
	def join_channel(self, u, c, **kwargs):
		"""Join a user to a channel, create channel if not exists"""
		
		use_mass = True
		if not isinstance(u, list):
			uid, user = self.get_user(u)
			if not user: return
			use_mass = False

		if self.burst:
			#optimizations, skips loop
			cname = c
			try:
				chan = self.channel_t[c]
			except KeyError:
				chan = None
		else:
			cname, chan = self.get_channel(c)

		if not chan:
			modes = kwargs.get('modes', "")
			limit = kwargs.get('limit', 0)
			key = kwargs.get('key', "")
			if use_mass: members = u
			else: members = [uid,]
			self.channel_t.update({c:{'modes':modes, 'limit':limit,
				'key':key, 'topic':'', 'members':members}})
		else:
			if use_mass: chan['members'].extend(u)
			else: chan['members'].append(uid)
			c = cname
		if use_mass:
			#speed hack
			user_t = self.user_t
			for uid in u:
				user_t[uid]['channels'].append(c)
		else: user['channels'].append(c)
	
	def part_channel(self, u, c):
		"""Part a user from a channel, remove channel if last user"""
		
		uid, user = self.get_user(u)
		if not user: return
		cname, chan = self.get_channel(c)
		if not chan: return
		
		try:
			user['channels'].remove(cname)
		except Exception, err:
			self.logger.log("Error, user wasn't on parted channel: %s" % err, L_ERROR)
		try:
			chan['members'].remove(uid)
		except Exception, err:
			self.logger.log("Error, channel list desynch detected (%s/%s): %s" % (uid,cname,err), L_ERROR)
		
		if len(chan['members']) == 0:
			try:
				del self.channel_t[cname]
			except Exception, err:
				self.logger.log("Error, channel disappeared during delete: %s" % err, L_ERROR)
	
	def quit_user(self, ident):
		"""Quits a user
		Remove all channels
		Remove from user_t
		"""
		
		uid, user = self.get_user(ident)
		if not user: return

		for c in user['channels'][:]:
			self.part_channel(uid,c)

		del self.user_t[uid]
			
	def is_id(self, id):
		"""Check if the supplied argument is a valid SID or UID"""
		
		if re.match("^[0-9][0-9A-Z]{2,8}$", id):
			return True
		else: return False
			
	def is_channel(self, name):
		"""Check if the supplied argument is a valid channel name"""
		
		if name.startswith("#"): return True
		else: return False
		
	def is_master(self, prefix):
		"""Check if the given user is in the master's list"""
		
		uid, user = self.get_user(prefix)
		if not user: return False

		if user['nick'] in self.masters and (
			config.getint('control', 'debug') or self.has_umode(uid, "r")):
			return True
		else: return False

			
	def has_umode(self, user, modes):
		"""Checks if a given user has usermode(s) set"""
		
		uid, user = self.get_user(user)
		if not user: return False

		for mode in modes:
			try:
				#TODO: would `if mode in user['modes']` work?
				if user['modes'].find(mode) == -1:
					return False
			except KeyError:
				return False
		return True
	
	def verify_acl(self, source, target, acl_dict):
		"""Check if a user has permission to execute a command per specified ACL
		"""
		
		if self.is_master(source): return True
		for acl_type, acl_args in acl_dict.iteritems():
			if acl_type == "channel":
				if target in acl_args: return True
			elif acl_type == "modes":
				if self.has_umode(source, acl_args): return True
			elif acl_type == "master": #redundant, but left for self-documentation
				if self.is_master(source): return True
		# it's either the callback/usage keys, or a typo
		return False
	
	def mkts(self):
		"""Return a unix timestamp as of now"""

		return int(time.time())

	def createFakeUser(self, nick, user, host, gecos, modes, nspass,
		flags = u"$=^_><|", version = u"PyPsd User", ip=u"0", away=True, *args, **kwargs):
		"""Introduce a new user into the network sourced from this server
		"""
		
		# convert ALL INPUT to unicode
		nick = self.c2u(nick)
		user = self.c2u(user)
		host = self.c2u(host)
		gecos = self.c2u(gecos)
		modes = self.c2u(modes)
		nspass = self.c2u(nspass)
		flags = self.c2u(flags)
		version = self.c2u(version)
		ip = self.c2u(ip)
		
		# if a user with that nickname exists, return the array for it.
		# this has the unintended feature of allowing multiple modules to share
		#  the same user if they try to add it.  failure happens however if one
		#  module unloads (and thus kills off the user), so this should not be
		#  attempted.
		xuid, xuser = self.get_user(nick)
		if xuser: return xuser

		uid = "%s%s" % (self.sid, self.base36encode(self.my_client_count))
		self.my_client_count += 1
		join_chans = kwargs.get('join_chans', [])
		join_chans.append(self.logchan)
		
		nu = {
			'nick':nick, 'user':user, 'host':host,
			'gecos':gecos, 'modes':modes, 'nspass':nspass,
			'version':version, 'ip':ip, 'rdns':host, 'server':self.hostname,
			'sid':self.sid, 'ts':u"%d" % self.mkts(),
			'flags':flags, 'su':None, 'channels':[]
		}
		
		self.user_t.update({uid:nu})
		
		self.sendMessage("UID", nick, "1", nu['ts'],
			modes, user, host, ip, uid,
			"0", host, ":%s" % gecos)
		self.sendMessage("ENCAP", "*", "AUTHFLAGS", nick,
			flags, prefix=self.hostname)
		self.privMsg('NickServ', "IDENTIFY %s" % nspass, sender=uid)
		for cname in join_chans:
			cname = self.c2u(cname)
			self.sendMessage("SJOIN", nu['ts'], cname,
				"+nt", "@%s" % uid, prefix=self.sid)
			self.join_channel(uid, cname)
		self.privMsg(self.logchan, "%s online (%s)" % (nick, self.version))
		if away:
			self.sendMessage("AWAY", ":I am a bot. email %s with questions" %
				config.get('control', 'master_email'), prefix=uid)
		self.logger.log("NEWUSER: %s" % nu, L_DEBUG)

		return (uid, nu)
	
	def quitFakeUser(self, uid):
		"""Quit off a user sourced from this network
		Could have been loaded from module, etc
		"""
		
		if uid[:3] == self.sid and uid in self.user_t:	
			self.quit_user(uid)
			self.sendMessage("QUIT", ":Module Unloaded", prefix=uid)
			self.logger.log("QUITUSER: %s" % uid, L_DEBUG)
			return True
		return False

	def killUser(self, target, sender, message):
		"""Issue a kill message for TARGET"""
		#TODO: can we do server kills?
		
		self.sendMessage("KILL", target, ": (%s)" % message, prefix=sender)
		
	def changeCmodes(self, cname, modes, *args):
		"""Changes modes of a channel in the local hash table
		target = channel name
		modes = (example, +rd-x)
		
		channel user modes can happen here.  for now, just ignore them.
		"""
		
		cname, chan = self.get_channel(cname)
		if not chan: return
			
		switchmode = None
		newmodes = chan['modes']
		ct = 0
		for mode in modes:
			if mode == "+" or mode == "-":
				switchmode = mode
			else:
				if switchmode == "+":
					if mode == "l":
						try:
							chan['limit'] = int(args[ct])
						except:
							chan['limit'] = 0
						ct += 1
					elif mode == "k":
						chan['key'] = args[ct]
						ct += 1
					elif mode in "vhoaqbeI":
						ct += 1
						continue
					if newmodes.count(mode) == 0:
						newmodes += mode
				elif switchmode == "-":
					if mode == "l":
						chan['limit'] = 0
					elif mode == "k":
						chan['key'] = ""
						ct += 1
					elif mode in "vhoaqbeI":
						ct += 1
						continue
					newmodes = newmodes.replace(mode, "")
		chan['modes'] = newmodes
		
	def changeUmodes(self, user, modes):
		"""Changes modes of a user in the local hash table
		target = UID or Nick
		modes = (example, +rd-x)
		"""
		
		uid, user = self.get_user(user)
		if not user: return
			
		switchmode = None
		newmodes = user['modes']
		for mode in modes:
			if mode == "+" or mode == "-":
				switchmode = mode
			else:
				if switchmode == "+" and newmodes.count(mode) == 0:
					newmodes += mode
				elif switchmode == "-":
					newmodes = newmodes.replace(mode, "")
		user['modes'] = newmodes
		
	def privMsg(self, target, message, sender=None):
		"""Send a message to a user or channel
		target can be a tuple or a list, in which case it
		issues the message to each in the list
		"""
		
		#if silence mode is enabled, don't print automatic ban notifications
		if self.silence and "Banned" in message: return

		if not sender: sender = self.uid

		if type(target).__name__ == "str":
			target = tuple((self.c2u(target),))
		elif type(target).__name__ == "unicode":
			target = tuple((target,))
		else: target = set(target)

		for dest in target:
			self.sendMessage("PRIVMSG", dest,
				":%s" % message, prefix=sender)

	def notice(self, target, message, sender=None):
		"""Send a notice to a user or channel"""

		if not sender: sender = self.uid
		self.sendMessage("NOTICE", target,
		":%s" % message, prefix=sender)

### End helper methods

### Begin command hooks
	def cmd_help(self, source, target, pieces):
		"""Prints a list of commands and syntax a user is allowed to use"""
		
		cmds = self.commands.keys()
		cmds.sort()
		for k in cmds:
			if self.verify_acl(source, target, self.commands[k]):
				self.notice(source,
					"%s%s %s" % (self.prefix, k, self.commands[k]['usage']))
		return True

	def cmd_lookup(self, source, target, pieces):
		"""Prints basic data about a user"""
		
		if not pieces: return False
		uid, mu = self.get_user(pieces[0])

		if not mu:
			self.privMsg(target, "User %s not found" % pieces[0])
			return True
		
		self.privMsg(target,
			"User (SU:%s) %s!%s@%s (%s), UID %s, Modes %s, Flags: %s, Server %s" %
			(mu['su'], mu['nick'], mu['user'], mu['host'], mu['gecos'],
			uid, mu['modes'], mu['flags'], mu['server']))
		return True
	
	def cmd_modload(self, source, target, pieces):
		"""Loads a specified module"""
		
		if not pieces: return False
			
		version = self.loadModule(pieces[0])
		if version:
			self.privMsg(target,
				"Module %s v%d loaded successfully" % (pieces[0], version))
		else: self.privMsg(target,
				"Module %s already loaded or failed to load" % pieces[0])
		return True
	
	def cmd_modunload(self, source, target, pieces):
		"""Unloads a specified module"""
		
		if not pieces: return False

		if self.unloadModule(pieces[0]):
			self.privMsg(target,
				"Module %s unloaded" % pieces[0])
		else:
			self.privMsg(target,
				"Module %s not loaded or failed shutdown()" % pieces[0])
		return True

	def cmd_modreload(self, source, target, pieces):
		"""Reloads a specified module"""

		if not pieces: return False

		if not self.reload(): self.privMsg(target, "Error reading config file for reloading")

		if self.cmd_modunload(source,target,pieces):
			self.cmd_modload(source,target,pieces)
		return True
	
	def cmd_reload(self, source, target, pieces):
		"""Re-read config file"""
		
		if self.reload():
			self.privMsg(target, "Reload successful")
		else:
			self.privMsg(target, "Reload failed, check your config file")
		return True
		
	def cmd_silence(self, source, target, pieces):
		"""Disable all automatic privmsg activity in logchan
		"""
		
		self.silence = not self.silence
		
		self.privMsg(target, "Silence enabled: %s" % self.silence)
		
		return True		

	def cmd_shutdown(self, source, target, pieces):
		"""Shuts down the system"""
		
		self.privMsg(target, "Shutting down...")
		self.shutdown()
		if len(pieces) > 0:
			message = " ".join(pieces)
		else: message = "shutting down"
		self.sendMessage("SQUIT", self.sid, ":%s" % message)
		return True
	
	def cmd_status(self, source, target, pieces):
		"""Displays runtime technical information"""
		tp = reactor.threadpool
		#print tp.dumpStats()
		#print tp.threads
		#print tp.waiters
		#print tp.working

		try:
			las = [str(x)[:4] for x in os.getloadavg()]
		except:
			las = ["unk", "unk", "unk"]

		self.privMsg(target, "Threads: %da+%di/%d Running; %d Queued (Load: [%s %s %s])" %
			(len(tp.working), len(tp.waiters), tp.max, len(tp.q.queue), las[0], las[1], las[2]))
		self.privMsg(target, "Modules loaded: %s" %
			(", ".join(self.modules)))
		return True
		
## End Command hooks

### Begin IRC hooks
	def irc_242(self, prefix, params):
		"""Receive a STATS U response"""
		
		self.logger.log('242/UPTIME: %s: %s' % (prefix, ' '.join(params)), L_IN)
		self.callHooks('242', prefix, params)
		
	def irc_439(self, prefix, params):
		"""Start the uplink negotiation
		A 439 signifies the "go ahead" for negotiation
		"""
		
		self.logger.log('Handshaking with uplink', L_STATUS)
		self.burst_timer = self.mkts()
		self.sendMessage("PASS",
			config.get('uplink', 'pass'), "TS", "6", ":%s" % self.sid)
		self.sendMessage("CAPAB",
			"TBURST TB UNKLN KLN GLN QS HUB ENCAP EOB EX IE CLUSTER CHW KNOCK")
			#topic burst x2(?), unkline, kline, gline, quitstorm, is a hub, encap, end of burst message, ex?, ie?, cluster?, chw?, knock?
		self.sendMessage("SERVER",
			self.hostname, "1", ":%s" % self.version)
		self.sendMessage("SVINFO",
			"6", "6", "%d" % self.mkts())
			
	def irc_BMASK(self, prefix, params):
		"""Called on burst, has data on:
		+b (bans)
		+e (ban exemptions)
		+I (invite exemptions)
		
		prefix = introducing server
		p0 = ts (of mode set)?
		p1 = chan name
		p2 = mode (e/I/b) (no +, not multiple)
		p3... = masks
		
		unimplemented currently
		"""
		
		pass
		
			
	def irc_ENCAP(self, prefix, params):
		"""Handle ENCAP messages
		We can ignore SVSJOIN/SVSPART since they trigger SJOIN/PART messages
		"""
		
		if params[1] == "SVSMODE":
			self.changeUmodes(params[2], params[4])
		elif params[1] == "AUTHFLAGS":
			uid, user = self.get_user(params[2])
			if user: user['flags'] = params[3]
		elif params[1] == "SU":
			uid, user = self.get_user(params[2])
			if not user: return
			if len(params) > 3:
				user['su'] = params[3]
			elif len(params) > 2:
				user['su'] = ""
		elif params[1] == "CHGHOST":
			uid, user = self.get_user(params[2])
			if user: user['host'] = params[3]
		self.callHooks('encap', prefix, params)
			
	def irc_EOB(self, prefix, params):
		"""Handle EOB (End of Burst) messages by disabling burst mode
		prefix = server EOB'd
		params = None
		"""
		
		self.burst = False
		self.privMsg(self.logchan,
			"Burst completed in %d seconds." %
			(self.mkts()-self.burst_timer))
			
	def irc_JOIN(self, prefix, params):
		"""Sent when a user joins an existing channel
		prefix = client UID
		p0 = ts
		p1 = channel name
		p2 = ?? (default modes?  always "+" in all dumps)
		"""
		
		self.join_channel(prefix, params[1])

	def irc_KICK(self, prefix, params):
		"""Remove user from channel on KICK
		prefix = kicker
		p0 = channel
		p1 = kicked
		p2... reason
		"""
		
		uid, user = self.get_user(params[1])
		if not user: return

		# if kicked, rejoin only if its from the log channel
		# this prevents the ban-evader problem
		if user['sid'] == self.sid and params[0] == self.logchan:
			self.sendMessage("JOIN", "%d" % self.mkts(), params[0], "+", prefix=uid)

		else:
			self.part_channel(uid, params[0])

	def irc_KILL(self, prefix, params):
		"""Handle received KILL messages
		Check if the user killed was one of our children, if so, resurrect them
		If not ours, remove them from the hash table
		
		prefix = killer
		p0 = killee
		p1... message
		"""
		
		uid, user = self.get_user(params[0])
		if not user: return

		self.callHooks('kill', prefix, params)
	
		if user['sid'] == self.sid:
			del self.user_t[uid]
			newuid, data = self.createFakeUser(**user)
			if uid == self.uid:
				self.uid = newuid
			else:
				for name, obj in self.modules.iteritems():
					if hasattr(obj, "uid") and uid == obj.uid:
						self.logger.log('RECREATE USER: OLD %s, NEW %s' % (uid, newuid), L_STATUS)
						obj.uid = newuid
		else:
			self.quit_user(params[0])		
	
	def irc_KLINE(self, prefix, params):
		"""Handle KLINE messages"""
		
		self.callHooks('kline', prefix, params)
			
	def irc_MODE(self, prefix, params):
		"""Handle mode changes
		prefix = user making the change
		p0 = channel or user
		p1 = modes
		p2... = mode arguments (for channels)
		"""
		
		if prefix == params[0]:
			self.changeUmodes(params[0], params[1])
		elif self.is_channel(params[0]):
			self.changeCmodes(params[0], params[1], *params[2:])
			
	def irc_NICK(self, prefix, params):
		"""Handle nick changes
		Remove mode +r, it'll get re-set by ENCAP SVSMODE if appropriate
		"""
		
		#cheaper to do a try/except than if
		try:
			self.user_t[prefix]['nick'] = params[0]
			self.changeUmodes(prefix, "-r")
		except Exception, err:
			self.logger.log("Error, unable to change nickname for %s: %s" % (prefix, err))
	
	def irc_NOTICE(self, prefix, params):
		"""Handle NOTICES received by the server
		Also check if the NOTICE was a CTCP and call that hook
		"""
		
		pieces = params[1].split(" ")
		if self.is_channel(params[0]): target = params[0]
		else: target = prefix
		if params[1].find("doing a whois") != -1:
			try:
				self.privMsg(self.logchan,
					"WHOIS: %s %s is doing a whois on me %s" %
					(pieces[3], pieces[4], pieces[11]), sender=params[0])
			except Exception, err:
				#Probably not a real whois line
				self.logger.log("Faked whois message received (%s/%s) (%s)" %
					(prefix,params,err), L_ERROR)
				
		if params[1][0] == "\001":
			self.callHooks('ctcp', prefix, params)
		else:
			self.callHooks('notice', prefix, params)
			
	def irc_PART(self, prefix, params):
		"""Sent when a user leaves a channel.
		Check to see if the channel should be destroyed (last user, etc)
		prefix = UID
		p0 = channel name
		"""
		
		self.part_channel(prefix, params[0])

	def irc_PASS(self, prefix, params):
		"""Handle PASS (password) messages from the uplink during negotiation.
		Check if the password is the one we have on file, disconnect if not.
		"""

		if self.burst:
			if params[0] != config.get('uplink', 'pass'):
				self.sendMessage("ERROR", ":Invalid downlink password")
				reactor.stop()
			self.uplink_sid = params[3]

	def irc_PING(self, prefix, params):
		"""Handle PING requests and send a proper pong."""

		if self.is_id(params[0]):
			sid, server = self.get_server(params[0])
			target = server['host']
		else: target = params[0]

		self.sendMessage("PONG", self.hostname, target, prefix=self.sid)
				
	def irc_PONG(self, prefix, params):
		"""Handle PONG replies, turn off bursting if we get one
		Typically a server only replies to PINGs after its EOB
		"""
		
		if self.burst:
			self.burst = False
			self.privMsg(self.logchan,
				"Burst completed in %d seconds." %
				(self.mkts()-self.burst_timer))
			
	def irc_PRIVMSG(self, prefix, params):
		"""Handle messages seen by the server.
		The command handler hook "hooks" in here
		"""
		
		uid, user = self.get_user(prefix)
		
		#throw in ctcp REQUESTS here, shouldn't be many
		if params[1][0] == "\001":
			muid, muser = self.get_user(params[0])
			if not muser: return

			if params[1][1:-1].upper() == 'VERSION':
				self.sendMessage("NOTICE", prefix, ":\001VERSION %s\001" %
					muser['version'], prefix=muid)
			return
		#handle privmsgs
		if params[1][0] != self.prefix:
		#	self.logger.log('PRIVMSG: %s: %s' % (user['nick'], '|'.join(params)), L_COMMAND)
			self.callHooks('privmsg', prefix, params)
			return
		
		self.logger.log('COMMAND: %s: %s' % (user['nick'], '|'.join(params)), L_COMMAND)
			
		pieces = params[1].strip().split(" ")
		if self.is_channel(params[0]):
			target = params[0]
		else: target = prefix
		
		cmd = pieces[0][1:]
		if cmd in self.commands:
			hook = self.commands[cmd]
			if self.verify_acl(prefix, target, hook):
				try:
					try: msg = pieces[1:]
					except: msg = None
					if not hook['callback'](prefix, target, msg):
						self.notice(prefix,
							"Usage: %s%s %s" % (self.prefix, cmd, hook['usage']))
				except Exception, err:
					self.logger.log('COMMAND: Error executing %s (%s)' %
						(cmd,err), L_ERROR)
					if config.getint('control', 'debug') == 1: raise
				return
				
		#self.privMsg(self.logchan,
		#	"SECURITY: %s attempted unauthorized command: %s" %
		#	(user['nick'], '|'.join(params)))
		self.logger.log("SECURITY: %s attempted unauthorized command: %s" %
			(user['nick'], '|'.join(params)), L_SECURITY)
		
	def irc_QUIT(self, prefix, params):
		"""Handle user quits from the network
		"""
		
		self.quit_user(prefix)
			
	def irc_SERVER(self, prefix, params):
		"""Handle SERVER messages during uplink negotiation
		SERVER messages ONLY appear during uplink negotiation AND for NON-TS6 SERVERS
		Since we don't give two craps about non-ts6, just ignore them if we don't see this
		 during negotiation/burst.
		"""
		
		if self.burst and self.first_burst:
			self.first_burst = False
			self.server_t.update({self.sid:{'host':self.hostname, 'hops':u"0",
				'desc':self.version, 'uplink':u"", 'uplink_sid':u""}})
			su = {
				'uplink':self.hostname, 'uplink_sid':self.sid,
				'host':params[0], 'hops':params[1], 'desc':params[2]
			}
			self.server_t.update({self.uplink_sid:su})
			self.server_tree.add_node(self.uplink_sid)
			self.server_tree.add_edge(self.sid, self.uplink_sid)
		
			if "netstats" in self.modules:
				self.sendMessage("VERSION", self.uplink_sid)
				
	def irc_SID(self, prefix, params):
		"""Handle SID (TS6 server introductions) messages
		prefix = Hub that links the server described in params
		"""
		
		if not self.burst:
			self.burst = True
			self.burst_timer = self.mkts()
			self.callHooks("sid", prefix, params)
			self.sendMessage("PING", self.sid, params[2])
		
		sid, server = self.get_server(prefix)
		
		su = {
			'uplink':server['host'], 'uplink_sid':sid,
			'host':params[0], 'hops':params[1], 'desc':params[3]
		}
		self.server_t.update({params[2]:su})
		self.server_tree.add_node(params[2])
		self.server_tree.add_edge(params[2], prefix)
		
		# netstats uses this
		if "netstats" in self.modules:
			self.sendMessage("VERSION", params[2])

	def irc_SJOIN(self, prefix, params):
		"""SJOIN's are server joins, used to relay new channel information.
		Often used during bursting and when new channels are made
		what clients are on what channels
		prefix = introducing server (SID)
		p0 = ts
		p1 = channel
		p2 = channel modes
		p[3...] = channel args (ugh)
		p[LAST] = user list, space seperated
		"""
		
		memlist = []
		offset = 0
		limit = 0
		key = ""
		modes = params[2].strip("0") # geoserv does weird stuff
		for mode in modes:
			if mode == "l":
				try:
					limit = int(params[3+offset])
				except:
					limit = 0
				offset += 1
			elif mode == "k":
				key = params[3+offset]
				offset += 1
		try:	
			self.join_channel(self.sjoinresub("", params[3+offset]).split(), params[1], limit=limit, key=key, modes=modes)
		except Exception, err:
			self.logger.log("Error in SJOIN: prefix=%s, params=%s, err=%s" % (prefix, "|".join(params), err), L_ERROR)
			
	def irc_SQUIT(self, prefix, params):
		"""Handle SQUIT (server quits)
		WE RECEIVE -ONE- SQUIT FOR HUB SPLITS
		 - determine who was on the other end
		 - delete off their users
		 - delete off the servers
		prefix = None
		params[0] = sid that squit
		params[1:] = squit message
		"""
		
		sid, server = self.get_server(params[0])
		if not server: return
			
		if self.sid == sid or self.uplink_sid == sid: return
			
		if self.burst: self.burst = False # in case a server never properly reconnects

		self.callHooks("squit", prefix, params)

		try:
			prune = nx.traversal.path.shortest_path(self.server_tree,
				self.sid, sid)[-2]
			self.server_tree.remove_edge(sid, prune)
			nodes = nx.strongly_connected_components(self.server_tree)
		except Exception, err:
			self.logger.log('Error, unable to tracepath SQUIT (%s)' % err, L_ERROR)
			self.privMsg(self.logchan, 'SQUIT CALCULATION ERROR')
			return
	
		if self.sid in nodes[0]:
			splits = nodes[1]
		else:
			splits = nodes[0]
		
		for uid, user in self.user_t.copy().iteritems():
			if user['sid'] in splits:
				self.quit_user(uid)

		for k in splits:
			del self.server_t[k]
			self.server_tree.remove_node(k)

	def irc_SVINFO(self, prefix, params):
		"""Handle SVSINFO command
		Things are pretty much cleared for burst start by this point, so start bursting
		Also load DEFAULT command handlers
		Also load modules too.
		Then send End Of Burst.
		"""
		
		self.uid, user = self.createFakeUser(
			config.get('control', 'nick'),
			config.get('control', 'user'),
			config.get('control', 'host'),
			config.get('control', 'gecos'),
			config.get('control', 'modes'),
			config.get('control', 'nspass')
		)
		self.registerCommand('modload',
			{'master':None, 'callback':self.cmd_modload,
			'usage':"<module name> - loads a module"})
		self.registerCommand('modunload',
			{'master':None, 'callback':self.cmd_modunload,
			'usage':"<module name> - unloads a module"})
		self.registerCommand('modreload',
			{'master':None, 'callback':self.cmd_modreload,
			'usage':"<module name> - reloads a module"})
		self.registerCommand('reload',
			{'master':None, 'callback':self.cmd_reload,
			'usage':"- reloads certain configuration parameters"})
		self.registerCommand('shutdown',
			{'master':None, 'callback':self.cmd_shutdown,
			'usage':"- shuts down the server and disconnects"})
		self.registerCommand('silence',
			{'channel':self.logchan, 'callback':self.cmd_silence,
			'usage':"- toggles visibility of ban messages to the log channel"})
		self.registerCommand('status',
			{'master':None, 'callback':self.cmd_status,
			'usage':"- display technical information"})
		self.registerCommand('lookup',
			{'channel':(self.logchan,), 'callback':self.cmd_lookup,
			'usage':"<nick> - lookup a user's uid, modes, and server"})
		self.registerCommand('help',
			{'modes':'o', 'callback':self.cmd_help,
			'usage':"- shows help for all commands available *to you*"})
		
		self.startModules()
		# let network know we are done bursting
		self.sendMessage("EOB", prefix=self.sid)
		self.sendMessage("PING", self.sid, self.uplink_sid)
		
	def irc_TBURST(self, prefix, params):
		"""TBURST's are topic bursts, which relay channel information
		during the burst phase.
		prefix = introducing server (NAME)
		p0 = ts (for channel creation?)
		p1 = channel name
		p2 = ts (for topic set?)
		(opt)p3 = topic setter (full nick!user@host)
		(opt)p4 = topic

		p3-4 do not exist if no topic.
		"""

		if len(params) < 3: return

		cname, chan = self.get_channel(params[1])
		if chan:
			try:
				chan['topic'] = params[4]
			except IndexError:
				self.logger.log("Error setting topic. params: %s" % ('|'.join(params)), L_ERROR)
	
	def irc_TMODE(self, prefix, params):
		"""Relay channel mode changes
		p0 = ts
		p1 = channel
		p2 = modes (-/+...)
		p3... = args
		
		Something to note here is that CHANNEL USER MODES happen here too!
		so +vvo user1 user2 user3 will propogate here (if we ever code in channel umodes)
		"""
		
		self.changeCmodes(params[1], params[2], *params[3:])
		self.callHooks("tmode", prefix, params)

	def irc_TOPIC(self, prefix, params):
		"""Relays TOPIC change information
		prefix = uid of channel changer
		p0 = channel name
		p1 = topic
		"""
		
		cname, chan = self.get_channel(params[0])
		if chan:
			chan['topic'] = params[1]
		
	def irc_UID(self, prefix, params):
		"""Called when a user connects to the network or is introduced.
		prefix = server they live on
		params:
		0 = nick
		1 = hops
		2 = TS
		3 = modes
		4 = username
		5 = spoof/nethost
		6 = realip (0 if spoofed!)
		7 = UID
		8 = ???
		9 = rdns
		10 = gecos

		Of this data, hops is useless to track, so throw it away
		"""
		sid, server = self.get_server(prefix)
		if not server:
			self.logger.log("Error, UID called from server %s, but it does not exist!(%s)" % (prefix, self.server_t), L_ERROR)
			return

		nu = {
			'nick':params[0], 'user':params[4], 'host':params[5],
			'gecos':params[10], 'modes':params[3],
			'version':"", 'ip':params[6], 'rdns':params[9],
			'server':server['host'], 'sid':prefix,
			'ts':params[2], 'flags':"", 'su':"", 'channels':[]
		}
		
		self.user_t.update({params[7]:nu})
		
		if self.flood and not self.flood_noticed:
			self.flood_noticed = True
			self.privMsg(self.logchan, "Flood detected, shutting down expensive systems...")
		elif not self.flood and self.flood_noticed:
			#only triggered on next user
			self.flood_noticed = False
			self.privMsg(self.logchan, "Flood gone, reactivating systems...")

		if not self.burst and params[6] != "0":
			#dont call hook on burst or an oper/spoofed
			self.callHooks("uid", prefix, params)

	def irc_UNKLINE(self, prefix, params):
		"""Handle UNKLINE messages"""
		
		self.callHooks('unkline', prefix, params)
	
	def irc_WHOIS(self, prefix, params):
		"""Handle WHOIS requests for server users.
		If requestor is an IRCOp, send them additional faked fields.
		"""
		
		muid, muser = self.get_user(params[0])
		ruid, ruser = self.get_user(prefix)

		if not muser:
			self.logger.log("Error, local whois user does not exist! (%s)" % params[0], L_ERROR)
			return

		mynick = muser['nick']
		self.privMsg(self.logchan,
			"WHOIS: %s is doing a whois on me [%s]" %
			(ruser['nick'], ruser['server']), sender=muid)
		self.sendMessage("311", prefix, "%s %s %s * :%s" %
			(mynick, muser['user'], muser['host'], muser['gecos']))
		#oper only
		if self.has_umode(prefix, "o"):
			self.sendMessage("319", prefix, mynick, ":%s" % " ".join(muser['channels']))
			self.sendMessage("312", prefix, mynick, self.hostname, self.version)
		else:
			self.sendMessage("312", prefix, mynick, "*.server", "hello there")
		self.sendMessage("313", prefix, mynick, ":is an IRC Operator - Network Services")
		#should make this real.
		self.sendMessage("307", prefix, mynick, ":has identified for this nick")
		#technically we are...
		self.sendMessage("671", prefix, mynick, ":is using a secure connection")
		#oper only
		if self.has_umode(prefix, "o"):
			self.sendMessage("310", prefix, mynick,
			":is using modes %s authflags: %s" %
			(muser['modes'], muser['flags']))
			self.sendMessage("338", prefix, mynick,
			":is actually %s@%s [%s]" %
			(muser['user'], muser['rdns'], "255.255.255.255"))
		self.sendMessage("318", prefix, mynick, ":End of WHOIS")
### End IRC hooks
